/*
Picsy v1.07   PIC16C84 programmer
Copyright (C) 1996-2000  Johan Dahlberg, Tomas Franzon

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Project: PICSY
  File:    hardware.c
  Version: 1.07

  This file contains hardware/operating system specific code for Picsy
*/



#ifdef __linux__
/*
  Created: 1996-11-01
  Changed: 2000-01-28
  Author:  Johan Dahlberg, johan_d@lysator.liu.se

  Operating system specific code for Linux
*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/serial.h>
#include "types.h"
#include "protos.h"


int fd = -1;      /* File descriptor for serial device */
ULONG uart;       /* Base address for the UART */
int line;         /* Current modem line status */
int dev_port_fd = -1; /* File descriptor for /dev/port */

#define IER	0x1	/* Interrupt enable register */
#define FCR	0x2	/* FIFO control register */
#define LCR	0x3	/* Line control register */
#define MCR	0x4	/* Modem control register */
#define MSR	0x6	/* Modem status register */



/* Code to read and write hardware registers.
   Unfortunately the serial driver doesn't seem
   to allow user programs to send infinite break
   signals.
*/

/* Open /dev/port
   Returns TRUE on success */
BOOL open_io()
{
  if ((dev_port_fd = open("/dev/port", O_RDWR)) < 0)
    return FALSE;
  return TRUE;
}

/* Close /dev/port */
void close_io()
{
  if (dev_port_fd >= 0) close(dev_port_fd);
  dev_port_fd = -1;
}

/* Write byte to port
   Returns TRUE on success */
BOOL port_out(int port, UBYTE d)
{
  if (lseek(dev_port_fd, port, 0) != port)
    return FALSE;
  
  if (write(dev_port_fd, &d, 1) != 1)
    return FALSE;

  return TRUE;
}

/* Read byte from port */
UBYTE port_in(int port)
{
  UBYTE d;

  if (lseek(dev_port_fd, port, 0) != port)
    return 0;

  if (read(dev_port_fd, &d, 1) != 1)
    return 0;

  return d;
}




/* Write some information about this hardware driver */
void hw_info(void)
{
  printf ("Linux-driver v1.03 by Johan Dahlberg\n");
}

/* How to use PICSY with this hardware driver */
void hw_usage(void)
{
  printf ("Usage:  PICSY erase   <chip> <device>\n");
  printf ("        PICSY program <chip> <hexfile> <device>\n");
  printf ("        PICSY verify  <chip> <hexfile> <device>\n");
  printf ("        PICSY read    <chip> <hexfile> <device>\n");
  printf ("        PICSY lock    <chip> <bits> <device>\n");
  printf ("        PICSY fuse    <chip> <bits> <device>\n\n");
  
  printf ("Where <chip> is name of chip (e.g. PIC16C84)\n");
  printf ("      <hexfile> is HEX-formatted file to read/write\n");
  printf ("      <bits> is a comma-separated list of lock/fuse bits\n"
	  "             to program\n"
	  "            (Try <bits>=help for list of supported bits)\n");
  printf ("      <device> is the serial device to use (e.g. /dev/ttyS1)\n");
}


/* Initialize programming hardware

  <string> points to a OS- and hardware specific
    configuration string (from command arguments)
    (may be ignored)

  Returns TRUE on success
*/
BOOL hw_init(char *string)
{
  struct serial_struct info;

  if (!*string) return FALSE;


  /* Here <string> tells us which device to use */
  if ((fd = open(string, O_RDWR|O_NDELAY)) <0) {
    perror (string);
    fprintf (stderr, "Unfortunately this porgram needs to be run with uid=root\n");
    return FALSE;
  }

  if (!open_io()) {
    fprintf (stderr, "Can't open /dev/port\n");
    fprintf (stderr, "Unfortunately this porgram needs to be run with uid=root\n");
    return FALSE;
  }



  /* Get UART base address */
  if (ioctl (fd, TIOCGSERIAL, &info)) {
    perror ("Can't get serial info\n");
    return FALSE;
  }
    
  uart = info.port;
  PDEBUG (("UART base address: 0x%lx\n", uart));
  hw_vppoff();
  hw_dataout(0);
  hw_clock(0);

  return TRUE;
}

/* Cleanup on exit
*/
void hw_cleanup(void)
{
  /* Initialize the serial port */
  hw_vppoff();
  hw_dataout(0);
  hw_clock(0);
  close_io();
  if (fd >= 0) close (fd);
}

/* Enable VCC and VPP to the processor
   (Set TxD-line on serial line HIGH)
*/
void hw_vppon(void)
{
  port_out(uart+LCR, 0x40);	/* Turn break on => TxD HIGH */
}

/* Remove VCC and VPP from the processor
   Set the MCLR-pin low (RESET active)
   (Set TxD-line on serial line LOW)
*/
void hw_vppoff(void)
{
  port_out(uart+LCR, 0);	/* Turn break off => TxD LOW */
}

/* Wait 100ns with VPP on */
void hw_waitus()
{
}

/* Wait n ms with VPP on */
void hw_waitms(UBYTE n)
{
  usleep (1000*(ULONG)n);
}

/* Put <level> on RB7 on the processor
   (Set DTR line HIGH if <level> is TRUE)
*/
void hw_dataout(BOOL level)
{
  const int dtr = TIOCM_DTR;
  
  ioctl (fd, level? TIOCMBIS:TIOCMBIC, &dtr);
}

/* Put <level> on RB6 on the processor
   (Set RTS line HIGH if <level> is TRUE)
*/
void hw_clock(BOOL level)
{
  const int rts = TIOCM_RTS;

  ioctl (fd, level? TIOCMBIS:TIOCMBIC, &rts);
}

/* Get level from RB7 on the processor
   (Return TRUE if CTS line is HIGH)
*/
BOOL hw_datain(void)
{
  int lines;

  ioctl (fd, TIOCMGET, &lines);
  return lines & TIOCM_CTS;
}
#else  /* Linux */





#ifdef __MSDOS__
/*
  Created: 1996-11-01
  Changed: 1996-11-13
  Author:  Johan Dahlberg, johan_d@lysator.liu.se

  Hardware specific code for i386/MSDOS
*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <bios.h>
#include <dos.h>
#include "Types.h"
#include "Protos.h"

UWORD uart; 	/* Base address for the UART */

#define IER	0x1	/* Interrupt enable register */
#define FCR	0x2	/* FIFO control register */
#define LCR	0x3	/* Line control register */
#define MCR	0x4	/* Modem control register */
#define MSR	0x6	/* Modem status register */

#define	COUNTER0 0x40	/* Time of day clock */
#define CONTROL  0x43	/* Timer mode control register */


/* Write some information about this hardware driver */
void hw_info(void)
{
  printf ("MSDOS-driver v1.03 by Johan Dahlberg\n");
}

/* How to use PICSY with this hardware driver */
void hw_usage(void)
{
  printf ("Usage:  PICSY erase   <chip> <comport>\n");
  printf ("        PICSY program <chip> <hexfile> <comport>\n");
  printf ("        PICSY verify  <chip> <hexfile> <comport>\n");
  printf ("        PICSY read    <chip> <hexfile> <comport>\n\n");
  printf ("        PICSY lock    <chip> <bits> <comport>\n");
  printf ("        PICSY fuse    <chip> <bits> <comport>\n\n");

  printf ("Where <chip> is name of chip (e.g. PIC16C84)\n");
  printf ("      <hexfile> is HEX-formatted file to read/write\n");
  printf ("      <bits> is a comma-separated list of lock/fuse bits\n"
	  "             to program\n"
	  "            (Try <bits>=HELP for list of supported bits)\n");
  printf ("      <comport> is the number of the COM-port to use (1-4)\n");
}


/* Initialize programming hardware

  <string> points to a OS- and hardware specific
    configuration string (from command arguments)
    (may be ignored)

  Returns TRUE on success
*/
BOOL hw_init(char *string)
{
  int comport;

  /* Here <string> tells us which COM-port to use */
  comport = atoi(string);
  if (!comport && !strlen(string)) comport = 1; /* Default: COM1 */
  if (comport < 1 || comport > 4) {
    fprintf (stderr, "COM%d not supported\n", comport);
    return FALSE;
  }

  /* Get the base address for the UART (from BIOS) */
  uart = *(UWORD far *)(0x400 + 2*(comport-1));

  /* Initialize the UART */
  outportb(uart+LCR, 0);	/* Turn break off */
  outportb(uart+FCR, 0x07);	/* Clear & disable FIFO */
  outportb(uart+IER, 0);	/* Disable all interrupts */
  outportb(uart+MCR, 0);	/* Turn DTR, RTS off */

  return TRUE;
}

/* Cleanup on exit
*/
void hw_cleanup(void)
{
  outportb(uart+LCR, 0x03);	/* Turn break off, set 8N1 */
  outportb(uart+FCR, 0xc7);	/* Clear & enable FIFO */
  outportb(uart+IER, 0);	/* Disable all interrupts */
  outportb(uart+MCR, 0);	/* Turn DTR, RTS off */
}

/* Enable VCC and VPP to the processor
   (Set TxD-line on serial line HIGH)
*/
void hw_vppon(void)
{
  outport(uart+LCR, 0x40);	/* Turn break on => TxD HIGH */
}

/* Remove VCC and VPP from the processor
   Set the MCLR-pin low (RESET active)
   (Set TxD-line on serial line LOW)
*/
void hw_vppoff(void)
{
  outport(uart+LCR, 0);		/* Turn break off => TxD LOW */
}

/* Wait 100ns with VPP on */
void hw_waitus()
{
}

/* Wait n ms with VPP on */
void hw_waitms(UBYTE n)
{
  delay(n);
}

/* Put <level> on RB7 on the processor
   (Set DTR line HIGH if <level> is TRUE)
*/
void hw_dataout(BOOL level)
{
  UBYTE data;

  data = inportb(uart+MCR);  /* Read register */
  data &= ~0x01;             /* Clear DTR bit */
  if (level) data |= 0x01;   /* Set DTR if level is TRUE */
  outportb(uart+MCR, data);  /* Write register */
}

/* Put <level> on RB6 on the processor
   (Set RTS line HIGH if <level> is TRUE)
*/
void hw_clock(BOOL level)
{
  UBYTE data;

  data = inportb(uart+MCR);  /* Read register */
  data &= ~0x02;             /* Clear RTS bit */
  if (level) data |= 0x02;   /* Set RTS if level is TRUE */
  outportb(uart+MCR, data);  /* Write register */
}

/* Get level from RB7 on the processor
   (Return TRUE if CTS line is HIGH)
*/
BOOL hw_datain(void)
{
  UBYTE data;

  data = inportb(uart+MSR);  /* Read register */
  return (data & 0x10);	     /* Mask out CTS bit */
}
#else  /* MSDOS */




#ifdef WIN32
/*
  Created: 1998-08-11
  Author:  Tomas Franzon, tomfr343@student.liu.se

  Hardware specific code for WIN32 (Windows 95/98/NT)
*/

#include <windows.h>
#include "Protos.h"
#include <stdio.h>

HANDLE hComm; 

/* Write some information about this hardware driver */
void hw_info(void)
{
  printf ("Win32-driver v1.0 by Tomas Franzon\n");
}

/* How to use PICSY with this hardware driver */
void hw_usage(void)
{
  printf ("Usage:  PICSY erase   <chip> <comport>\n");
  printf ("        PICSY program <chip> <hexfile> <comport>\n");
  printf ("        PICSY verify  <chip> <hexfile> <comport>\n");
  printf ("        PICSY read    <chip> <hexfile> <comport>\n");
  printf ("        PICSY lock    <chip> <bits> <comport>\n");
  printf ("        PICSY fuse    <chip> <bits> <comport>\n\n");

  printf ("Where <chip> is name of chip (e.g. PIC16C84\n");
  printf ("      <hexfile> is HEX-formatted file to read/write\n");
  printf ("      <bits> is a comma-separated list of lock/fuse bits\n"
	  "             to program\n"
	  "            (Try <bits>=HELP for list of supported bits)\n");
  printf ("      <comport> is the number of the COM-port to use (1-4)\n");
}


/* Initialize programming hardware

  <string> points to a OS- and hardware specific
    configuration string (from command arguments)
    (may be ignored)

  Returns TRUE on success
*/
BOOL hw_init(char *string)
{
  char Port[50];
  DWORD  EvtMask;

	if (string[0]=='\0')
		strcpy(string,"1");
	strcpy(Port,"COM");
	strcat(Port,string);
	hComm = CreateFile(Port, GENERIC_READ,0,0,OPEN_EXISTING,0,0);
	if (hComm==INVALID_HANDLE_VALUE)
	{
		if (GetLastError()==5) 
		  printf("Cant't open %s! The com port is used by an other application.\r\n",Port);
		else
		  printf("Can't open %s! The com port is not installd on this machine.\r\n",Port);
		return(FALSE);
	}
	EvtMask=EV_CTS;
	SetCommMask(hComm,EvtMask);
	hw_vppoff();

  return TRUE;
}

/* Cleanup on exit
*/
void hw_cleanup(void)
{
	if (hComm!=INVALID_HANDLE_VALUE)
	{
	  hw_vppoff();
    CloseHandle(hComm);
	}
}

/* Enable VCC and VPP to the processor
   (Set TxD-line on serial line HIGH)
*/
void hw_vppon(void)
{
  SetCommBreak(hComm);	/* Turn break on => TxD HIGH */
}

/* Remove VCC and VPP from the processor
   Set the MCLR-pin low (RESET active)
   (Set TxD-line on serial line LOW)
*/
void hw_vppoff(void)
{
  ClearCommBreak(hComm);		/* Turn break off => TxD LOW */
}

/* Wait 100ns with VPP on */
#pragma optimize("",off)
void hw_waitus()
{
}
#pragma optimize("",on)

/* Wait n ms with VPP on */
void hw_waitms(UBYTE n)
{
	/* 10ms is to little, 20ms works better...?!?! */
  Sleep(n*2);
}

/* Put <level> on RB7 on the processor
   (Set DTR line HIGH if <level> is TRUE)
*/
void hw_dataout(BOOL level)
{
	if (level)
		EscapeCommFunction(hComm,SETDTR);
	else
		EscapeCommFunction(hComm,CLRDTR);
}

/* Put <level> on RB6 on the processor
   (Set RTS line HIGH if <level> is TRUE)
*/
void hw_clock(BOOL level)
{
	if (level)
		EscapeCommFunction(hComm,SETRTS);
	else
		EscapeCommFunction(hComm,CLRRTS);
}

/* Get level from RB7 on the processor
   (Return TRUE if CTS line is HIGH)
*/
BOOL hw_datain(void)
{
	DWORD  ModemStat;
	GetCommModemStatus(hComm,&ModemStat);
	return (MS_CTS_ON&ModemStat);
}
#else  /* WIN32 */
#error "Your computer architecture/OS is currently not supported\n"
#endif /* WIN32 */
#endif /* MSDOS */
#endif /* Linux */
