/*
Picsy v1.07   PIC and AVR programmer
Copyright (C) 1996-2000  Johan Dahlberg

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


  Project: PICSY
  File:    Hex.c
  Version: 1.07
  Created: 1996-10-25
  Changed: 2000-01-24
  Author:  Johan Dahlberg, johan_d@lysator.liu.se

  Hex-file reader / writer
*/

#include <stdio.h>
#include <stdlib.h>
#include "types.h"
#include "protos.h"
#include "devices.h"


#define LINEBUF_LEN 256
char linebuf[LINEBUF_LEN];
int  linepos;
UBYTE checksum;


/* Clear the program buffer */
void clearbuf (devtype *device)
{
  ULONG address;
  meminfo *info;
  int count;

  for (count = 0; count < device->Blocks; ++count) {
    info = &device->MemInfo[count];

    for (address=info->FirstAddr; address<=info->LastAddr; ++address) {
      buffer[address] = address&1 ? info->DefaultOdd : info->DefaultEven;
    }
  }

  for (count=0; count<(BUFFER_SIZE/16); ++count)
    buf_flag[count] = FALSE;
}


/* Convert an ASCII-HEX char to binary
*/
UBYTE char2hex (char c)
{
  switch (c) {
  case '0': return 0;
  case '1': return 1;
  case '2': return 2;
  case '3': return 3;
  case '4': return 4;
  case '5': return 5;
  case '6': return 6;
  case '7': return 7;
  case '8': return 8;
  case '9': return 9;
  case 'a':
  case 'A': return 10;
  case 'b':
  case 'B': return 11;
  case 'c':
  case 'C': return 12;
  case 'd':
  case 'D': return 13;
  case 'e':
  case 'E': return 14;
  case 'f':
  case 'F': return 15;
  }

  fprintf (stderr, "Input file not a HEX-file\n");
  exit(10);
}



/* Read two chars from 'linebuf' and convert from ASCII-HEX
   to binary. Calculate checksum.
*/
UBYTE getbyte (void)
{
  UBYTE data;

  data = 16*char2hex(linebuf[linepos++]);
  data += char2hex(linebuf[linepos++]);
  checksum += data;

  return data;
}


/* Read a HEX-file to the buffer */
BOOL readhex (char *filename, devtype *device, BOOL extended)
{
  FILE	*file;
  UBYTE	data;
  ULONG address, maxaddr = 0;
  int i;

  for (i=0; i<device->Blocks; ++i) {
    if (device->MemInfo[i].FirstAddr >= 0x10000) {
      if (!extended) continue;
    }
    else if (extended) continue;
    if (maxaddr < device->MemInfo[i].LastAddr) maxaddr = device->MemInfo[i].LastAddr;
  }
  ++maxaddr;
  

  printf ("Reading HEX-file...\n");
  file = fopen(filename, "r");
  if (!file){
    fprintf (stderr, "File not found: %s\n", filename);
    return FALSE;
  }

  /*--- Scan the HEX-file ---*/
  while (fgets(linebuf, LINEBUF_LEN, file)) {
    if (linebuf[0] != ':') continue;   /* Beginning of a HEX-file line? */
    linepos = 1;
    checksum = 0;

    data = getbyte();         /* How many databytes on this line? */
    address = 256*getbyte();  /* Get address to put them into */
    address += getbyte();
    if (extended) address += 0x10000;
    if ((address+data) > maxaddr) {
	fprintf (stderr, "HEX file address outside chip space\n");
	fclose (file);
	return FALSE;
    }
    getbyte();                /* I'm not sure what this byte really does... */

    /* Read data to program buffer */
    for (; data; --data) {
      buf_flag[address/16] = TRUE;
      buffer[address++] = getbyte();
    }

    /* Get and test checksum */
    getbyte();
    if (checksum) {
      fprintf (stderr, "Input file not a HEX-file\n");
      fclose(file);
      return FALSE;
    }

  }

  fclose (file);

  return TRUE;
}


/* Write a block of data to a HEX-file */
BOOL writeblock(FILE *file, ULONG start, ULONG end)
{
  char *ptr;
  int count;
  UBYTE checksum;

  while (start <= end) {
    count = (end-start)+1;
    if (count > 16) count = 16;

    sprintf (linebuf, ":%02X%04lX00", count, start%65536);
    checksum = 0-(count + (start/256)%256 + start%256);
    ptr = &linebuf[9];

    for (; count; --count) {
      sprintf(ptr, "%02X", buffer[start]);
      ptr += 2;
      checksum -= buffer[start++];
    }

    sprintf(ptr, "%02X\n", checksum);
    if (fputs (linebuf, file) < 0) {
      return FALSE;
    }
  }

  return TRUE;
}


/* Write the buffer to a HEX-file */
BOOL writehex(char *filename, devtype *device, BOOL extended)
{
  FILE *file;
  int i;
  BOOL skip;

  printf ("Writing HEX-file...\n");
  /* Create file for output */
  file = fopen(filename, "wt");
  if (!file) {
    fprintf (stderr, "Can't create file: %s\n", filename);
    return FALSE;
  }

  i = 0;
  while (TRUE) {
    /* Save memory block(s) */
    skip = FALSE;
    if (device->MemInfo[i].FirstAddr >= 0x10000) {
      if (!extended) skip = TRUE;
    }
    else if (extended) skip = TRUE;

    if (!skip) {
      if (!writeblock(file, device->MemInfo[i].FirstAddr, device->MemInfo[i].LastAddr))
	break;
    }

    ++i;
    if (i < device->Blocks) continue;

    /* Save end-of-file block */
    if (fputs(":00000001FF\n", file) < 0)
      break;

    fclose(file);

    return TRUE;  /* Success! */
  }

  fprintf (stderr, "Error during file write\n");
  fclose(file);

  return FALSE;
}


/* Dump buffer contents to screen */
void dump(void)
{
  UWORD address;
  int i;

  for (address = 0x4000; address < 0x4300;) {
    printf ("%04X: ", address);

    for (i = 0; i < 16; ++i) {
      printf ("%02X ", buffer[address++]);
      if (i%4 == 3) printf (" ");
    }
    printf ("\n");
  } 
}
