/*
Picsy v1.07   PIC and AVR programmer
Copyright (C) 1996-2000  Johan Dahlberg

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


  Project: PICSY
  File:    Picsy.c
  Version: 1.07
  Created: 1996-10-25
  Changed: 2000-01-27
  Author:  Johan Dahlberg, johan_d@lysator.liu.se

  Software for PIC16x84/AVR/I2C EEPROM programmer
*/

#define DEV_INFO
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "types.h"
#include "protos.h"
#include "devices.h"

UBYTE buffer[BUFFER_SIZE];      /* Pic program data buffer */
UBYTE buf_flag[BUFFER_SIZE/16]; /* Pic program data buffer modification flags */

/* Display program usage */
void usage(void)
{
  printf ("Picsy comes with absolutely NO WARRANTY.\n");
  printf ("This software is free and may be distributed under the terms of\n");
  printf ("GNU Public License.\n\n");
  hw_usage();
  exit (10);
}

/* Display a list of supported chips */
void supported_chips(void)
{
  int i;
  devtype *chip;

  fprintf (stderr, "Currently supported chips are:");
  for (i=0;; ++i) {
    if (!(i%4)) fprintf(stderr, "\n");
    chip = &dev_types[i];
    if (!chip->Name) break;
    fprintf (stderr, " %-19s", chip->Name);
  }
  fprintf(stderr, "\n");
}


/* Well.... Main */
#define CMD_ERASE   1
#define CMD_PROGRAM 2
#define CMD_VERIFY  3
#define CMD_READ    4
#define CMD_LOCK    5
#define CMD_FUSE    6
int main (int argc, char **argv)
{
  int args = 3;
  int cmd = 0;
  char *filename;
  char *hwopts;
  devtype *device = &dev_types[0];

  printf ("PICSY - Admittansen PIC/AVR/EEPROM programmer version 1.07 by Johan Dahlberg\n");
  hw_info();
  printf ("\n");

  /* Get command */
  if (argc < 3) usage();
  if (strucmp(argv[1], "ERASE"))   cmd = CMD_ERASE;
  if (strucmp(argv[1], "PROGRAM")) cmd = CMD_PROGRAM;
  if (strucmp(argv[1], "VERIFY"))  cmd = CMD_VERIFY;
  if (strucmp(argv[1], "READ"))    cmd = CMD_READ;
  if (strucmp(argv[1], "LOCK"))    cmd = CMD_LOCK;
  if (strucmp(argv[1], "FUSE"))    cmd = CMD_FUSE;
  if (!cmd) usage();

  /* Select the correct chip type */
  while (1) {
    if (!device->Name) {
      fprintf (stderr, "Unknown chip: %s\n\n", argv[2]);
      supported_chips();
      return 10;
    }
    if (strucmp (argv[2], device->Name)) break;
    ++device;
  }

  /* Get hardware options */
  if (cmd != CMD_ERASE) {
    if (argc < 4) usage();
    filename = argv[3];
    args = 4;
  }
  hwopts = (argc > args) ? argv[args]:"";



  /* Initialize the programmer hardware */
  if (!hw_init(hwopts)) {
    fprintf (stderr, "Hardware initialization failure\n");
    return 10;
  }
  atexit(hw_cleanup);

  /* Perform selected command */
  clearbuf(device);
  return device->Handler(cmd, filename, device);
}


int pic_handler (UBYTE cmd, char *filename, devtype *device)
{
  switch (cmd) {
  case CMD_ERASE:
    if (!pic_erase(device)) return 5;
    break;

  case CMD_PROGRAM:
    if (!readhex(filename, device, FALSE)) return 10;
    if (!strucmp(device->Name, "PIC16C84")) {
      /* Erase FLASH-chips before programming */
      if (!pic_erase(device)) return 5;
    }
    if (!pic_program(device)) return 5;

    if (!(buffer[0x400e] & 0x10)) break; /* Code protection enabled? */
    if (!pic_verify(device)) return 5;
    break;

  case CMD_VERIFY:
    if (!readhex(filename, device, FALSE)) return 10;
    if (!pic_verify(device)) return 5;
    break;

  case CMD_READ:
    if (!pic_read(device)) return 5;
    if (!writehex(filename, device, FALSE)) return 10;
    break;

  case CMD_LOCK:
    fprintf(stderr, "LOCK command currently not supported\n");
    return 10;

  case CMD_FUSE:
    fprintf(stderr, "FUSE command not supported by device family\n");
    return 10;
  }
  printf ("OK\n");

  return 0;
}


int avr_handler (UBYTE cmd, char *filename, devtype *device)
{
  char eepname[256]; /* Name of eeprom data file */
  char *s;
  FILE *f;

  /* Check if there's any *.eep file (containing EEPROM data) */
  strncpy(eepname, filename, 251);
  eepname[255] = 0;
  s = strchr(eepname, '.');
  if (s) {
    strcpy (s, ".eep");

    if (cmd != CMD_READ) {
      f = fopen(eepname, "r");
      if (!f) s = NULL;
      else fclose(f);
    }
  }
  
  
  switch (cmd) {
  case CMD_ERASE:
    if (!avr_erase(device)) return 5;
    break;

  case CMD_PROGRAM:
    if (!readhex(filename, device, FALSE)) return 10;
    if (s) {
      if (!readhex(eepname, device, TRUE)) return 10;      
    }
    if (!avr_erase(device)) return 5;
    if (!avr_program(device)) return 5;
    if (!avr_verify(device)) return 5;
    break;

  case CMD_VERIFY:
    if (!readhex(filename, device, FALSE)) return 10;
    if (s) {
      if (!readhex(eepname, device, TRUE)) return 10;      
    }
    if (!avr_verify(device)) return 5;
    break;

  case CMD_READ:
    if (!avr_read(device)) return 5;
    if (!writehex(filename, device, FALSE)) return 10;
    if (s) {
      if (!writehex(eepname, device, TRUE)) return 10;      
    }
    break;

  case CMD_LOCK:
    if (!avr_lock(filename, device)) return 5;
    break;

  case CMD_FUSE:
    if (!avr_fuse(filename, device)) return 5;
    break;
  }
  printf ("OK\n");

  return 0;
}


int eep_handler (UBYTE cmd, char *filename, devtype *device)
{
  switch (cmd) {
  case CMD_ERASE:
    if (!eep_erase(device)) return 5;
    break;

  case CMD_PROGRAM:
    if (!readhex(filename, device, FALSE)) return 10;
    if (!eep_program(device)) return 5;
    if (!eep_verify(device)) return 5;
    break;

  case CMD_VERIFY:
    if (!readhex(filename, device, FALSE)) return 10;
    if (!eep_verify(device)) return 5;
    break;

  case CMD_READ:
    if (!eep_read(device)) return 5;
    if (!writehex(filename, device, FALSE)) return 10;
    break;

  case CMD_LOCK:
    fprintf(stderr, "LOCK command not supported by device family\n");
    return 10;

  case CMD_FUSE:
    fprintf(stderr, "FUSE command not supported by device family\n");
    return 10;
  }
  printf ("OK\n");

  return 0;
}
