/*
Picsy v1.07   PIC and AVR programmer
Copyright (C) 1996-2000  Johan Dahlberg

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


  Project: PICSY
  File:    Program.c
  Version: 1.07
  Created: 1996-11-01
  Changed: 2000-01-24
  Author:  Johan Dahlberg, johan_d@lysator.liu.se

  Reads/writes to the PIC/AVR/EEPROM using the functions in Hardware.c
*/

#include <stdio.h>
#include <string.h>
#include "types.h"
#include "protos.h"
#include "devices.h"

/*-------------------------------------------------------------------------*/
/*                         PIC processor programming                       */
/*-------------------------------------------------------------------------*/


BOOL enablevpp(void)
{
  int i;

  for (i=0; i < 10; ++i) {
    hw_dataout(FALSE);   /* Data & clock lines low */
    hw_clock(FALSE);
    hw_vppon();	         /* Enable VCC and VPP */
    hw_waitms(10);
    hw_vppoff();	/* Remove VPP => perform reset */
    hw_waitus();
    hw_vppon();	        /* Enable VPP */
    hw_waitms(1);


    if (testmode()) return TRUE;;
  }

  fprintf (stderr, "Error initializing the chip\n");
  return FALSE;
}


/* Program the processor memories */
BOOL pic_program(devtype *device)
{
  meminfo *info;
  ULONG address;
  UWORD data;
  UBYTE cmdr, cmdw;
  int i, retry, block;

  /* Test the code protect bit
     and erase the device if necessary */
  if (!enablevpp()) return FALSE;
  sendcmd(PCMD_LOADCFG);
  writedata(0);
  for (i=0; i<7; ++i)
    sendcmd(PCMD_INCADDR);
  sendcmd(PCMD_READPGM);
  data = readdata();
  PDEBUG (("Config data: 0x%x\n", data));
  hw_vppoff();
  hw_waitms(1);

  if (!(data & 0x10)) pic_erase(device); /* Code protection enabled! */


  for (block=0; block<device->Blocks; ++block) {
    info = &device->MemInfo[block];
    address = info->FirstAddr;

    if (!enablevpp()) return FALSE;
    switch(info->UserData) {
    case AREA_PROGRAM:
      printf ("Writing program memory...\n");
      cmdw = PCMD_LOADPGM;
      cmdr = PCMD_READPGM;
      break;

    case AREA_DATA:
      printf ("Writing data memory...\n");
      cmdw = PCMD_LOADDATA;
      cmdr = PCMD_READDATA;
      break;

    case AREA_ID:
      printf ("Writing ID bits...\n");
      cmdw = PCMD_LOADPGM;
      cmdr = PCMD_READPGM;
      sendcmd(PCMD_LOADCFG);
      writedata(0);
      break;

    case AREA_CONFIG:
      printf ("Writing configuration word...\n");
      cmdw = PCMD_LOADPGM;
      cmdr = PCMD_READPGM;
      sendcmd(PCMD_LOADCFG);
      writedata(0);

      for (address = 0x4000; address<info->FirstAddr; address+=2) {
	sendcmd(PCMD_INCADDR);
      }
    }

    while (address <= info->LastAddr) {
      //PDEBUG (("Programming address 0x%lx\n", address/2));
      data = buffer[address++];
      data += 256*buffer[address++];

      /* Do we have to program this word? */
      if (!buf_flag[(address-2)/16]) {
	sendcmd(PCMD_INCADDR);
	continue;
      }

      /* Does the memory cell already contain correct data? */
      sendcmd(cmdr);
      if (!((readdata()^data) & info->Mask)) {
	sendcmd(PCMD_INCADDR);
	continue;
      }

      for (retry=0;; ++retry) {
	/* Start programming cycle */
	sendcmd(cmdw);  /* Write data to memory */
	writedata(data);
	sendcmd(PCMD_PROGRAM); /* Begin programming */
	hw_waitms(10);
	
	/* Code protection ON? */
	if ((address == 0x4010) && !(data & 0x10)) break;
	
	/* Verify memory contents after write */
	sendcmd(cmdr);     /* Verify memory contents */
	if (!((readdata()^data) & info->Mask)) break;
	if (retry<5) continue; /* Retry! */
	
	hw_vppoff();
	printf ("Programming failure at address 0x%X\n", (address-2)/2);
	return FALSE;
      }
      sendcmd(PCMD_INCADDR); /* Increment address */
    }
    hw_vppoff();
    hw_waitms(1);
  }

  return TRUE;
}

/* Verify the processor memories */
BOOL pic_verify(devtype *device)
{
  meminfo *info;
  ULONG address;
  UWORD data;
  UBYTE cmdr;
  int i, retry, block;


  for (block=0; block<device->Blocks; ++block) {
    info = &device->MemInfo[block];
    address = info->FirstAddr;

    if (!enablevpp()) return FALSE;
    switch(info->UserData) {
    case AREA_PROGRAM:
      printf ("Verifying program memory...\n");
      cmdr = PCMD_READPGM;
      break;

    case AREA_DATA:
      printf ("Verifying data memory...\n");
      cmdr = PCMD_READDATA;
      break;

    case AREA_ID:
      printf ("Verifying ID bits...\n");
      cmdr = PCMD_READPGM;
      sendcmd(PCMD_LOADCFG);
      writedata(0);
      break;

    case AREA_CONFIG:
      printf ("Verifying configuration word...\n");
      cmdr = PCMD_READPGM;
      sendcmd(PCMD_LOADCFG);
      writedata(0);

      for (address = 0x4000; address<info->FirstAddr; address+=2) {
	sendcmd(PCMD_INCADDR);
      }
    }

    while (address <= info->LastAddr) {
      //PDEBUG (("Programming address 0x%lx\n", address/2));
      data = buffer[address++];
      data += 256*buffer[address++];

      /* Do we have to verify this word? */
      if (!buf_flag[(address-2)/16]) {
	sendcmd(PCMD_INCADDR);
	continue;
      }

      /* Does the memory cell contain correct data? */
      sendcmd(cmdr);
      if (!((readdata()^data) & info->Mask)) {
	sendcmd(PCMD_INCADDR);
	continue;
      }

      hw_vppoff();
      printf ("Verify failure at address 0x%X\n", (address-2)/2);
      return FALSE;
    }
    hw_vppoff();
    hw_waitms(1);
  }

  return TRUE;
}


/* Read the processor memories */
BOOL pic_read(devtype *device)
{
  meminfo *info;
  ULONG address;
  UWORD data;
  UBYTE cmdr;
  int i, retry, block;


  for (block=0; block<device->Blocks; ++block) {
    info = &device->MemInfo[block];
    address = info->FirstAddr;

    if (!enablevpp()) return FALSE;
    switch(info->UserData) {
    case AREA_PROGRAM:
      printf ("Reading program memory...\n");
      cmdr = PCMD_READPGM;
      break;

    case AREA_DATA:
      printf ("Reading data memory...\n");
      cmdr = PCMD_READDATA;
      break;

    case AREA_ID:
      printf ("Reading ID bits...\n");
      cmdr = PCMD_READPGM;
      sendcmd(PCMD_LOADCFG);
      writedata(0);
      break;

    case AREA_CONFIG:
      printf ("Reading configuration word...\n");
      cmdr = PCMD_READPGM;
      sendcmd(PCMD_LOADCFG);
      writedata(0);

      for (address = 0x4000; address<info->FirstAddr; address+=2) {
	sendcmd(PCMD_INCADDR);
      }
    }

    while (address <= info->LastAddr) {
      sendcmd(cmdr);  /* Read data from memory */
      data = readdata();
      buffer[address++] = data%256;
      buffer[address++] = data/256;
      sendcmd(PCMD_INCADDR); /* Increment address */
    }
    hw_vppoff();
    hw_waitms(1);
  }

  return TRUE;
}


/* Clear the whole processor, including config word */
BOOL pic_erase(devtype *device)
{
  int i;

  if (!enablevpp()) return FALSE;

  printf ("Erasing the entire processor...\n");
  sendcmd(PCMD_LOADCFG);
  writedata(0x10);
  for (i=0; i<7; ++i)
    sendcmd(PCMD_INCADDR);
  sendcmd(PCMD_ERASE1);
  sendcmd(PCMD_ERASE2);
  sendcmd(PCMD_PROGRAM);
  hw_waitms(10);
  sendcmd(PCMD_ERASE1);
  sendcmd(PCMD_ERASE2);

  hw_vppoff();
  hw_waitms(1);
  return TRUE;
}


/* Send command to the processor */
void sendcmd(UBYTE cmd)
{
  int i;

  for (i=0; i<6; ++i) {
    hw_dataout(cmd & 0x01);  /* Transmit command, LSB first */
    cmd >>= 1;
    hw_waitus();
    hw_clock(TRUE);
    hw_waitus();
    hw_clock(FALSE);
    hw_waitus();
  }
}

/* Write data to the processor */
void writedata(UWORD data)
{
  int i;

  data &= 0x3fff;  /* Make sure bit 14 is cleared */

  hw_dataout(FALSE);  /* Start bit */
  hw_waitus();
  hw_clock(TRUE);
  hw_waitus();
  hw_clock(FALSE);
  hw_waitus();

  for (i=0; i<15; ++i) {
    hw_dataout(data & 0x0001);  /* Transmit 14 databits + stopbit, LSB first */
    data >>= 1;
    hw_waitus();
    hw_clock(TRUE);
    hw_waitus();
    hw_clock(FALSE);
    hw_waitus();
  }
}

/* Read data from the processor */
UWORD readdata(void)
{
  UWORD data = 0;
  int i;

  hw_dataout(FALSE);  /* Start bit */
  hw_waitus();
  hw_clock(TRUE);
  hw_waitus();
  hw_clock(FALSE);
  hw_waitus();
  hw_dataout(TRUE);

  for (i=0; i<14; ++i) {  /* Read 14 databits, LSB first */
    hw_waitus();
    hw_clock(TRUE);
    data >>= 1;
    hw_waitus();
    hw_clock(FALSE);
    data |= hw_datain() ? 0x2000:0x0000;
  }


  hw_waitus();
  hw_clock(TRUE);   /* Stop bit */
  hw_waitus();
  hw_clock(FALSE);
  hw_dataout(FALSE);

  return data;
}


/* Test if the processor has entered test mode */
BOOL testmode(void)
{
  int i;
  BOOL data;

  /* Send read program command */
  sendcmd (PCMD_READPGM);

  hw_dataout(FALSE);  /* Start bit */
  hw_waitus();
  hw_clock(TRUE);
  hw_waitus();
  hw_clock(FALSE);
  hw_waitus();

  for (i=0; i<14; ++i) {  /* Read 14 databits, LSB first */
    hw_dataout(TRUE);
    hw_waitus();
    hw_clock(TRUE);
    hw_waitus();
    hw_clock(FALSE);
    hw_waitus();
    data = hw_datain();

    hw_dataout(FALSE);
    hw_waitus();
    if (data != hw_datain())
      return FALSE;
  }


  hw_waitus();
  hw_clock(TRUE);   /* Stop bit */
  hw_waitus();
  hw_clock(FALSE);
  hw_dataout(FALSE);

  return TRUE;  /* Processor has entered test mode */
}



/*-------------------------------------------------------------------------*/
/*                   AVR processor device programming                      */
/*-------------------------------------------------------------------------*/

/* Write command and read response */
ULONG avr_cmd(ULONG dout)
{
  ULONG din = 0;
  int i;

  /* Write and read 8 bits */
  for (i=0; i<32; ++i) {
    hw_dataout((dout & 0x80000000)? 1:0);
    dout <<= 1;
    din <<= 1;
    hw_waitus();
    hw_waitus();
    hw_clock(1);
    din |= hw_datain()? 0x00000001:0x00000000;
    hw_waitus();
    hw_clock(0);
  }

  return din;
}


/* Make the AVR go into program mode */
BOOL avr_init (void)
{
  int retries = 32;

  hw_vppoff();
  hw_clock(0);
  hw_dataout(0);
  hw_waitms(20); /* Make sure the SCK line is low */

  hw_vppon(); /* Short pulse on the /RESET line */
  hw_waitms(1);
  hw_vppoff();
  hw_waitms(20); /* Wait at least 20ms */

  while (retries) {
    if ((avr_cmd(ACMD_PGMEN) & 0x0000ff00) == 0x00005300) break;
    --retries;

    hw_clock(1); /* Pulse SCK line */
    hw_waitus();
    hw_clock(0);
    hw_waitus();
  }
  if (!retries) {
    fprintf (stderr, "No functional AVR device connected\n");
    return FALSE;
  }

  return TRUE;
}


/* Make the AVR identify itself */
BOOL avr_identify(devtype *device)
{
  ULONG signature, d;
  flaginfo f;
  int i;

  /* Read signature */
  d = avr_cmd(ACMD_RDDEV|0x0000) & 0xff; /* Read manufacturer code */
  signature = d << 16;
  d = avr_cmd(ACMD_RDDEV|0x0100) & 0xff; /* Read memory size and type */
  signature |= d << 8;
  d = avr_cmd(ACMD_RDDEV|0x0200) & 0xff; /* Read controller stepping */
  signature |= d;


  /* Correct chip selected? */
  f.Name = NULL;
  f.Help = NULL;
  f.Value = signature;
  f.Group = AGRP_DEVID;
  if (flag_find (device->Flags, &f)) return TRUE;
  
  /* Find out what device it might be */
  for (device = dev_types; device;) {
    device = device_find(device, &f);
    if (!device) break;

    if ((void*)device->Handler != (void*)avr_handler) {
      /* Not an AVR chip */
      ++device;
      if (!device->Name) break; /* End of chip list */
      continue;
    }

    /* AVR device with matching signature found */
    fprintf (stderr, "Incorrect chip type\n");
    fprintf (stderr, "The chip is probably a %s\n", device->Name);
    return FALSE;
  }


  fprintf (stderr, "Protected chip or incorrect chip type\n"
	           "Erase the device to remove lock bits\n" );
  if ((signature & 0xff0000) == 0x1e0000) {
    fprintf (stderr, "\nThe chip is an AVR device with signature 0x%04x\n"
	     "Please download the latest program version. And - if you still\n"
	     "have problems with your chip - send an email with the chip name\n"
	     "(e.g. AT90S8515) and the signature above to johan_d@lysator.liu.se\n",
	     signature & 0xffff);
  }
  return FALSE;
}


/* Program the processor */
BOOL avr_program(devtype *device)
{
  meminfo *info;
  ULONG address;
  int block;
  ULONG wcmd, rcmd;

  if (!avr_init()) return FALSE;
  if (!avr_identify(device)) return FALSE;

  /* Write all memory blocks */
  for (block=0; block<device->Blocks; ++block) {
    info = &device->MemInfo[block];
    address = info->FirstAddr;
    switch (info->UserData) {
    case AREA_PROGRAM:
      printf ("Writing program memory...\n");
      break;

    case AREA_DATA:
      printf ("Writing data memory...\n");
      break;
    }

    while (address <= info->LastAddr) {
      /* Skip if nothing read from file to these addresses */
      if (!buf_flag[address/16]) {
	address += 16;
	continue;
      }

      /* Read/write commands to use */
      switch (info->UserData) {
      case AREA_PROGRAM:
	wcmd = ACMD_WRPMEM;
	rcmd = ACMD_RDPMEM;
	wcmd |= (address&0xfffe)<<7;
	rcmd |= (address&0xfffe)<<7;
	if (address & 0x01) {
	  wcmd |= 0x08000000;
	  rcmd |= 0x08000000;
	}
	break;
	
      case AREA_DATA:
	wcmd = ACMD_WREMEM;
	rcmd = ACMD_RDEMEM;
	wcmd |= (address&0xffff)<<8;
	rcmd |= (address&0xffff)<<8;
	break;
      }
      wcmd |= buffer[address];

      if ((avr_cmd(rcmd) & 0xff) != buffer[address]) {
	/* Start write */
	avr_cmd(wcmd);
	
	/* Wait for write to complete */
	if ((buffer[address] == 0x7f)||(info->UserData == AREA_DATA)) {
	  hw_waitms(4);
	}
	else {
	  while ((avr_cmd(rcmd) & 0xff) == 0x7f)
	    ;
	}
	
	/* Verify */
	if ((avr_cmd(rcmd) & 0xff) != buffer[address]) {
	  fprintf (stderr, "Programming error at address 0x%lx\n", address%0x10000);
	  return FALSE;
	}
      }
      ++address;
    }
  }

  return TRUE;
}


/* Verify the programming */
BOOL avr_verify(devtype *device)
{
  meminfo *info;
  ULONG address;
  int block;
  ULONG rcmd;

  if (!avr_init()) return FALSE;
  if (!avr_identify(device)) return FALSE;

  /* Verify all memory blocks */
  for (block=0; block<device->Blocks; ++block) {
    info = &device->MemInfo[block];
    address = info->FirstAddr;
    switch (info->UserData) {
    case AREA_PROGRAM:
      printf ("Verifying program memory...\n");
      break;

    case AREA_DATA:
      printf ("Verifying data memory...\n");
      break;
    }

    while (address <= info->LastAddr) {
      /* Skip if nothing read from file to these addresses */
      if (!buf_flag[address/16]) {
	address += 16;
	continue;
      }

      /* Read command to use */
      switch (info->UserData) {
      case AREA_PROGRAM:
	rcmd = ACMD_RDPMEM;
	rcmd |= (address&0xfffe)<<7;
	if (address & 0x01) {
	  rcmd |= 0x08000000;
	}
	break;
	
      case AREA_DATA:
	rcmd = ACMD_RDEMEM;
	rcmd |= (address&0xffff)<<8;
	break;
      }
      
      /* Verify */
      if ((avr_cmd(rcmd) & 0xff) != buffer[address]) {
	fprintf (stderr, "Verify error at address 0x%lx\n", address);
	return FALSE;
      }
      ++address;
    }
  }

  return TRUE;
}


/* Read data from the AVR processor */
BOOL avr_read(devtype *device)
{
  meminfo *info;
  ULONG address;
  int block;
  ULONG rcmd;

  if (!avr_init()) return FALSE;
  if (!avr_identify(device)) return FALSE;

  /* Read all memory blocks */
  for (block=0; block<device->Blocks; ++block) {
    info = &device->MemInfo[block];
    address = info->FirstAddr;
    switch (info->UserData) {
    case AREA_PROGRAM:
      printf ("Reading program memory...\n");
      break;

    case AREA_DATA:
      printf ("Reading data memory...\n");
      break;
    }

    while (address <= info->LastAddr) {
      /* Read command to use */
      switch (info->UserData) {
      case AREA_PROGRAM:
	rcmd = ACMD_RDPMEM;
	rcmd |= (address&0xfffe)<<7;
	if (address & 0x01) {
	  rcmd |= 0x08000000;
	}
	break;
	
      case AREA_DATA:
	rcmd = ACMD_RDEMEM;
	rcmd |= (address&0xffff)<<8;
	break;
      }

      /* Read byte */
      buffer[address] = avr_cmd (rcmd) & 0xff;
      ++address;
    }
  }

  return TRUE;
}


/* Erase the entire processor */
BOOL avr_erase(devtype *device)
{
  if (!avr_init()) return FALSE;

  printf ("Erasing the entire processor\n");
  avr_cmd(ACMD_ERASE);

  avr_init();
  avr_identify(device);
  
  return TRUE;
}


/* Write lock bits */
BOOL avr_lock(char *mode, devtype *device)
{
  flaginfo template;
  flaginfo *f;
  ULONG lockbits = 0xffffffff;
  char *name;

  template.Help  = NULL;
  template.Value = 0;
  template.Group = AGRP_LOCK;

  /* Parse all fuse names in comma-separated mode string */
  while (name = strtok(mode, ",")) {
    mode = NULL;
    template.Name = name;
    
    /* Find binary value for lockbit name */
    if (!(f = flag_find(device->Flags, &template))) {
      /* Fuse name not known */
      template.Name = NULL;
      fprintf (stderr, "%s supports the following lock bits:\n", device->Name);
      if (!flag_list (stderr, device->Flags, &template)) {
	fprintf (stderr, " None\n");
      }
      return FALSE;
    }
    else {
      lockbits &= f->Value;
    }
  }
  
  if (!avr_init()) return FALSE;
  if (!avr_identify(device)) return FALSE;

  printf ("Writing lock bits\n");
  avr_cmd(ACMD_WRLOCK & lockbits);
  return TRUE;
}


/* Write fuse bits */
BOOL avr_fuse(char *mode, devtype *device)
{
  flaginfo template;
  flaginfo *f;
  ULONG fusebits = 0xffffffff;
  char *name;

  template.Help  = NULL;
  template.Value = 0;
  template.Group = AGRP_FUSE;

  /* Parse all fuse names in comma-separated mode string */
  while (name = strtok(mode, ",")) {
    mode = NULL;
    template.Name = name;
    
    /* Find binary value for fuse name */
    if (!(f = flag_find(device->Flags, &template))) {
      /* Fuse name not known */
      template.Name = NULL;
      fprintf (stderr, "%s supports the following fuse bits:\n", device->Name);
      if (!flag_list (stderr, device->Flags, &template)) {
	fprintf (stderr, " None\n");
      }
      return FALSE;
    }
    else {
      fusebits &= f->Value;
    }
  }
  
  if (!avr_init()) return FALSE;
  if (!avr_identify(device)) return FALSE;

  printf ("Writing fuse bits\n");
  avr_cmd(ACMD_WRFUSE & fusebits);
  return TRUE;
}


/*-------------------------------------------------------------------------*/
/*                         Serial EEPROM programming                       */
/*-------------------------------------------------------------------------*/



/* Send I2C start condition */
void i2c_start (void)
{
  hw_clock(0);
  hw_waitus();
  hw_dataout(1);
  hw_waitus();
  hw_clock(1);
  hw_waitus();
  hw_dataout(0);
  hw_waitus();
  hw_clock(0);
}

/* Send I2C stop condition */
void i2c_stop (void)
{
  int i;

  for (i=0; i<1; ++i) {
    hw_clock(0);
    hw_waitus();
    hw_dataout(0);
    hw_waitus();
    hw_clock(1);
    hw_waitus();
    hw_dataout(1);
    hw_waitus();
    hw_clock(0);
  }
}

/* Send byte on I2C bus, returns TRUE on success */
BOOL i2c_putbyte (UBYTE data)
{
  int i;

  /* Send 8 databits */
  hw_dataout(0);
  for (i=0; i<8; ++i) {
    hw_dataout(data&0x80);
    hw_waitus();
    hw_clock(1);
    hw_waitus();
    hw_clock(0);
    hw_waitus();
    data <<= 1;
  }

  /* Get acknowledge */
  hw_dataout(1);
  hw_waitus();
  hw_clock(1);
  hw_waitus();
  i = hw_datain();
  hw_clock(0);

  return !i;
}


/* Get byte from I2C bus */
UBYTE i2c_getbyte(BOOL doack)
{
  UBYTE data = 0;
  int i;

  hw_dataout(1);
  
  /* Get 8 databits */
  for (i=0; i<8; ++i) {
    data <<= 1;
    hw_waitus();
    hw_clock(1);
    hw_waitus();
    if (hw_datain()) data |= 0x01;
    hw_clock(0);
  }

  /* Send acknowledge */
  if (doack) hw_dataout(0);
  hw_waitus();
  hw_clock(1);
  hw_waitus();
  hw_clock(0);

  return data;
}


/* Write data to EEPROM */
BOOL eep_program(devtype *device)
{
  meminfo *info;
  ULONG address;
  UBYTE chip;
  int retrycount;

  info = device->MemInfo;
  address=info->FirstAddr;

  printf ("Writing EEPROM memory...\n");
  hw_vppon();
  hw_dataout(1);
  hw_clock(1);
  hw_waitms(20);


  while(address<=info->LastAddr) {
    /* Set current address */
    i2c_start();
    chip = 0xa0;
    if (info->LastAddr > 2047) {
      /* Memory >= 32Kbit */
      if (!i2c_putbyte(chip)) {
	printf ("No device present!\n");
	hw_vppoff();
	return FALSE;
      }
      if (!i2c_putbyte(address/256)) {
	printf ("Error writing memory\n");
	hw_vppoff();
	return FALSE;
      }
    }
    else {
      /* Memory < 32Kbit */
      chip |= (address/128) & 0x0e;
      if (!i2c_putbyte(chip)) {
	printf ("No device present!\n");
	hw_vppoff();
	return FALSE;
      }
    }
    if (!i2c_putbyte(address%256)) {
      printf ("Error writing memory\n");
      hw_vppoff();
      return FALSE;
    }
      
    /* Write data to memory */
    while(address<=info->LastAddr) {
      if (!i2c_putbyte(buffer[address])) {
	printf ("Error writing memory\n");
	hw_vppoff();
	return FALSE;
      }
      ++address;
      if (!(address&0x0003)) break; /* Write 4 bytes at a time */
    }

    hw_waitms(1);
    i2c_stop();

    /* Wait for write cycle to complete */
    for (retrycount=1000; retrycount; --retrycount) {
      hw_waitms(2);
      i2c_start();
      if (i2c_putbyte(chip)) break;
      i2c_stop();
    }
    i2c_stop();
    if (!retrycount) {
      printf ("Error writing memory\n");
      hw_vppoff();
      return FALSE;
    }
  }
  i2c_stop();
  hw_vppoff();

  return TRUE;
}

/* Verify data in EEPROM */
BOOL eep_verify(devtype *device)
{
  meminfo *info;
  ULONG address;
  UBYTE chip = 0xa0;

  info = device->MemInfo;
  address=info->FirstAddr;

  printf ("Verifying EEPROM memory...\n");
  hw_vppon();
  hw_dataout(1);
  hw_clock(1);
  hw_waitms(20);

  /* Set current address */
  i2c_start();
  if (info->LastAddr > 2047) {
    /* Memory >= 32Kbit */
    if (!i2c_putbyte(chip)) {
      printf ("No device present!\n");
      hw_vppoff();
      return FALSE;
    }
    if (!i2c_putbyte(address/256)) {
      printf ("Error reading memory\n");
      hw_vppoff();
      return FALSE;
    }
  }
  else {
    /* Memory < 32Kbit */
    chip |= (address/128) & 0x0e;
    if (!i2c_putbyte(chip)) {
      printf ("No device present!\n");
      hw_vppoff();
      return FALSE;
    }
  }

  if (!i2c_putbyte(address%256)) {
    printf ("Error reading memory\n");
    hw_vppoff();
    return FALSE;
  }
  
  /* Send read command */
  i2c_start();
  if (!i2c_putbyte(chip|0x01))  {
    printf ("Error reading memory\n");
    hw_vppoff();
    return FALSE;
  }
  

  /* Read data */
  for (; address<info->LastAddr; ++address) {
    if (buffer[address] != i2c_getbyte(TRUE)) {
      printf ("Verify failure at address 0x%X\n", address);
      hw_vppoff();
      return FALSE;
    }
  }
  if (buffer[address] != i2c_getbyte(FALSE)) {
    printf ("Verify failure at address 0x%X\n", address);
    hw_vppoff();
    return FALSE;
  }
  i2c_stop();

  hw_vppoff();
  return TRUE;
}


/* Read data from EEPROM */
BOOL eep_read(devtype *device)
{
  meminfo *info;
  ULONG address;
  UBYTE chip = 0xa0;

  info = device->MemInfo;
  address=info->FirstAddr;

  printf ("Reading EEPROM memory...\n");
  hw_vppon();
  hw_dataout(1);
  hw_clock(1);
  hw_waitms(20);

  /* Set current address */
  i2c_start();
  if (info->LastAddr > 2047) {
    /* Memory >= 32Kbit */
    if (!i2c_putbyte(chip)) {
      printf ("No device present!\n");
      hw_vppoff();
      return FALSE;
    }
    if (!i2c_putbyte(address/256)) {
      printf ("Error reading memory\n");
      hw_vppoff();
      return FALSE;
    }
  }
  else {
    /* Memory < 32Kbit */
    chip |= (address/128) & 0x0e;
    if (!i2c_putbyte(chip)) {
      printf ("No device present!\n");
      hw_vppoff();
      return FALSE;
    }
  }

  if (!i2c_putbyte(address%256)) {
    printf ("Error reading memory\n");
    hw_vppoff();
    return FALSE;
  }
  
  /* Send read command */
  i2c_start();
  if (!i2c_putbyte(chip|0x01))  {
    printf ("Error reading memory\n");
    hw_vppoff();
    return FALSE;
  }
  

  /* Read data */
  for (; address<info->LastAddr; ++address) {
    buffer[address] = i2c_getbyte(TRUE);
  }
  buffer[address] = i2c_getbyte(FALSE);
  i2c_stop();

  hw_vppoff();
  return TRUE;
}

/* Erase the EEPROM */
BOOL eep_erase(devtype *device)
{
  return eep_program(device);
}
